//
// C++ Interface: Animation
//
// Description: 
//
//
// Author: Francois Faure <Francois.Faure@imag.fr>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef Animation_h
#define Animation_h
class SoSeparator;
class SoGroup;
class SoTransform;
class SoCoordinate3;
class SoIndexedFaceSet;
class SoDrawStyle;
class SoCamera;
class SoCylinder;
class SbMatrix;

#include <vector>
#include <Inventor/fields/SoMFVec3f.h>


#include <Skinned.h>
#include <subdivision.h>
#include <Skeleton.h>
#include <double_vector.h>

                 
class Animation{
public:



    Animation( SoGroup* );

    ~Animation();
    
    /// Update DOF values
    void update( float t, float dt);

    int set_camera(SoCamera* _camera);
    




    protected:





    //FUNCTIONS
    //*********//

    //build multiresolution in time
    int build_time_multires(const std::vector <double> angle,std::vector <double_vector> *average,std::vector<double_vector> *residual);

    //load the mesh and subdivide it
    int load_meshes(const char* filename,SoGroup *parent);

    //load the skeleton from collada file
    int load_skeleton(const char* filename,const char* name_first_node);

    //set manually animation: sum of sinus
    int set_animation(int k_bone,int k_dim,double A,double B,double C,double omega_1,double omega_2,double omega_3);
    
    //associate skeleton to graph nodes
    int create_skeleton_nodes(SoGroup* parent);

    //build threshold for subdivision
    int build_distance_threshold();


    //DATA
    //*********//

    //number of subdivision in space/time
    int N_subdiv,N_time_subdiv;

    std::vector <Skinned> skinned_vector; //set of skinned mesh
    std::vector <double> space_threshold; //the threshold in space
    std::vector <double> time_threshold;  //the threshold in time

    std::vector <SoSeparator*> separator_meshes;       //graph separator to add meshs
    std::vector <SoCoordinate3*> coordinates_meshes;   //coordinates in graph
    std::vector <SoIndexedFaceSet*> index_face_meshes; //face index in graph
    std::vector <SoDrawStyle*> style_meshes;           //drawing style in graph

    std::vector <Skeleton> skeleton_vector; //set of skeleton (N_time_subdiv)

    std::vector <SoSeparator*> separator_skeleton; //separator for skeleton in graph
    std::vector <SoCylinder*> cylinder_vector;     //cylinder in graph (to draw skeleton)
    std::vector <SoTransform*> skeleton_transform; //cylinder transform



    Skeleton *skeleton; 
    Skeleton *deformed_skeleton;
    SoCamera *camera;



};

#endif
