
#ifndef   	MESH_H_
# define   	MESH_H_

#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string>
#include <errno.h>
#include <stdio.h>

#include <vector>
#include <list>
using namespace std;



#include <V_3D.h>
#include <Matrix.h>
#include <int_vector.h>

class Mesh
{
 public:

  //constructors
  Mesh();
  Mesh(const Mesh &);
  ~Mesh();

  int destroy();
  int destroy_connectivity();

  //vertex
  double& get_vertex(int k_vertex,int k_dim);//get vertex form its index
  double& get_vertex(int k_triangle,int k_vertex,int k_dim); //from k_triangle
  double get_vertex(int k_vertex,int k_dim) const;
  double get_vertex(int k_triangle,int k_vertex,int k_dim) const; 
  V_3D get_vertex_v3d(int k_vertex) const; //get vertex from its index
  V_3D get_vertex_v3d(int k_triangle,int k_vertex) const; //vertex
							  //from
							  //k_triangle
  int set_vertex(int k_vertex,double x,double y,double z);
  int set_vertex(int k_vertex,int k_dim,double value);
  int set_vertex(int k_vertex,V_3D V);
  int set_vertex_add(int k_vertex,int k_dim,double d_value);
  int set_vertex_add(int k_vertex,const V_3D& d_value);
  int get_vertex_size() const;
  int get_vertex_number() const;

  int add_vertex(double value);
  int add_vertex(double x,double y,double z);
  int add_vertex(V_3D v);

  int rotate(const V_3D& n,double angle);
  int rotate(double n_x,double n_y,double n_z,double angle);
  int rotate(double *n,double angle);
  int translate(double t_x,double t_y,double t_z);
  int translate(double *t);
  int scale(double s_x,double s_y,double s_z);
  int scale(double s);
  int scale(double *s);
  int get_closest_vertex(double x,double y,double z);


  //index vertex = triangle connectivity
  int& get_index_vertex(int k_index,int k_dim);
  int get_index_vertex(int k_index,int k_dim) const;
  int& get_triangle_number(int k_triangle,int k_dim);
  int get_triangle_number(int k_triangle,int k_dim) const;
  int set_index_vertex(int k_triangle_index,int k_vertex,double value);
  int get_index_vertex_size() const;
  int get_triangle_number() const;

  int add_triangle(int index_0,int index_1,int index_2);
  int add_triangle(std::vector <int> index);
  int add_polygon(const std::vector <int>& index);

  int reverse_triangle_order();

  int load_connectivity(const Mesh& _mesh);
  
  
  //normal
  double& get_normal(int k_normal,int k_dim);
  double& get_normal(int k_triangle,int k_vertex,int k_dim);
  double get_normal(int k_normal,int k_dim) const;
  double get_normal(int k_triangle,int k_vertex,int k_dim) const;
  V_3D get_normal_v3d(int k_normal) const;
  V_3D get_normal_v3d(int k_triangle,int k_vertex) const;
  int set_normal(int k_normal,int k_dim,double value);
  int set_normal(int k_normal,V_3D V);
  int build_smooth_normal();
  int is_color() const;


  //color
  double& get_color(int k_color,int k_dim);
  double& get_color(int k_triangle,int k_vertex,int k_dim);
  double get_color(int k_color,int k_dim) const;
  double get_color(int k_triangle,int k_vertex,int k_dim) const;
  int set_color(int k_color,int k_dim,double value);
  int fill_color(double r,double g,double b,double alpha);
  int fill_color(double r,double g,double b);
  int fill_alpha(double alpha);
  int increase_alpha(double d_alpha);



  //check
  int fast_check_integrity();

  int unitize_size();


  int build_fast_visualization_pointers();
  int is_builded_pointers();
  double *get_p_vertex();
  double *get_p_normal();
  double *get_p_color();
  int *get_p_index_vertex();


  Mesh& operator=(const Mesh&);

  V_3D get_middle_point() const;
  V_3D get_min_point() const;
  V_3D get_max_point() const;
  V_3D get_size_mesh() const;

  int build_one_ring();
  int get_neighbors_number(int k_vertex) const;
  int get_neighbors(int k_vertex,int k_neighbors) const;
  int set_neighbors(int k_vertex,int k_neighbors,int index);
  int is_neighbors_two_ring(int k_vertex,int k_other_vertex);
  int is_neighbors(int k_vertex,int index);
  int get_neighbors_two_ring_number(int k_vertex) const;
  int get_neighbors_two_ring(int k_vertex,int k_neighbors) const;
  std::vector <int_vector> get_one_ring();
  std::vector <int_vector> get_two_ring();
  
  int smooth_surface();
  int subdivide_surface();
  int laplace_scheme(Mesh *mesh2,const std::vector <int>& linked_index,int N_new_vertex);


  int cut_plane(double a,double b,double c,double d);//cut with plane
						     //ax+by+cz+d=0
                                                     //direction of normal n=(a,b,c)
  int cut_plane(const V_3D& n,const V_3D& x0);
  int cut_plane(const V_3D& n,double d);
  int select_box(const V_3D& p1,const V_3D& p2);

  int keep_triangles(int tri_start,int tri_end);
  int delete_triangles(int tri_start,int tri_end);

  double norm_2(const Mesh& _mesh) const;

  std::vector<int> build_circle(int N_decompose);
  std::vector<int> build_square(int N_1,int N_2);



  const double* get_vertex();
  const double* get_normal();
  const double* get_color();
  const int* get_index_vertex();


  //adjacent faces (triangulation only)
  int build_adjacent_faces();
  int get_adjacent_faces(int k_face,int k_edge) const;



  std::vector <int> get_connectivity_vrml();


 private:

  int empty_pointers();

  int create_linked_index_if_needed(int index_0,int index_1,std::vector <int> *linked_index,int *count);
  int build_boundary_edge(std::list <int>* boundary,int index_0,int index_1,int current_index);
  int is_boundary(const std::list <int>& boundary,int k_index);
  
  int find_opposite_face(int current_face,int edge_start,int edge_end) const;




  std::vector <double> vertex;
  std::vector <double> normal;
  std::vector <double> color;
  std::vector <int> index_vertex;


  std::vector <int_vector> one_ring;
  std::vector <int_vector> two_ring;

  std::vector <int> adjacent_faces;
  
  //fast visualization pointers
  int is_fast_visualization_pointers;
  double *p_vertex;
  double *p_normal;
  double *p_color;
  int *p_index_vertex;


};



#endif 	    /* !MESH_H_ */
