
#include <Mesh.h>


Mesh::Mesh():
  vertex(),normal(),color(),index_vertex(),one_ring(),two_ring(),adjacent_faces(),is_fast_visualization_pointers(0),p_vertex(NULL),p_normal(NULL),p_color(NULL),p_index_vertex(NULL)
{
  vertex.resize(0);
  normal.resize(0);
  color.resize(0);
  index_vertex.resize(0);
  one_ring.resize(0);
  two_ring.resize(0);
  adjacent_faces.resize(0);

}


Mesh::Mesh(const Mesh &_M):
  vertex(),normal(),color(),index_vertex(),one_ring(),two_ring(),adjacent_faces(),is_fast_visualization_pointers(0),p_vertex(NULL),p_normal(NULL),p_color(NULL),p_index_vertex(NULL)
{

  unsigned int k=0;
  
  vertex.resize(_M.vertex.size());
  for(k=0;k<vertex.size();k++)
    vertex[k]=_M.vertex[k];

  index_vertex.resize(_M.index_vertex.size());
  for(k=0;k<index_vertex.size();k++)
    index_vertex[k]=_M.index_vertex[k];

  normal.resize(_M.normal.size());
  for(k=0;k<normal.size();k++)
    normal[k]=_M.normal[k];

  color.resize(_M.color.size());
  for(k=0;k<color.size();k++)
    color[k]=_M.color[k];

  one_ring.resize(_M.one_ring.size());
  for(k=0;k<one_ring.size();k++)
    one_ring[k]=_M.one_ring[k];

  two_ring.resize(_M.two_ring.size());
  for(k=0;k<two_ring.size();k++)
    two_ring[k]=_M.two_ring[k];

  adjacent_faces.resize(_M.adjacent_faces.size());
  for(k=0;k<adjacent_faces.size();k++)
    adjacent_faces[k]=_M.adjacent_faces[k];
  
  

  is_fast_visualization_pointers=0;
  p_vertex=NULL;
  p_normal=NULL;
  p_color=NULL;
  p_index_vertex=NULL;
    
}

Mesh::~Mesh()
{
  destroy();
}


int Mesh::destroy()
{

  vertex.resize(0);
  index_vertex.resize(0);
  normal.resize(0);
  color.resize(0);

  one_ring.resize(0);
  two_ring.resize(0);
  adjacent_faces.resize(0);

  empty_pointers();


  return 0;
}

int Mesh::destroy_connectivity()
{
  index_vertex.resize(0);
  empty_pointers();
  return 0;
}


double& Mesh::get_vertex(int k_vertex,int k_dim)
{
  
  if(3*k_vertex+k_dim<0 || 3*k_vertex+k_dim>=int(vertex.size()))
    {
      printf("Error in get_vertex in Mesh, k_vertex[%d,%d] is not correct for size [%d]\n",k_vertex,k_dim,vertex.size());
      exit(-1);
    }
    
    return vertex[3*k_vertex+k_dim];
}

double Mesh::get_vertex(int k_vertex,int k_dim) const
{
  
  if(3*k_vertex+k_dim<0 || 3*k_vertex+k_dim>=int(vertex.size()))
    {
      printf("Error in get_vertex in Mesh, k_vertex[%d,%d] is not correct for size [%d]\n",k_vertex,k_dim,vertex.size());
      exit(-1);
    }
    
    return vertex[3*k_vertex+k_dim];
}

int Mesh::set_vertex(int k_vertex,int k_dim,double value)
{
  if(3*k_vertex+k_dim<0 || 3*k_vertex+k_dim>=int(vertex.size()))
    {
      printf("Error in set_vertex in Mesh, k_vertex[%d,%d] is not correct for size [%d]\n",k_vertex,k_dim,vertex.size());
      exit(-1);
    }

    vertex[3*k_vertex+k_dim]=value;
    return 0;
}

int Mesh::set_vertex(int k_vertex,double x,double y,double z)
{
  set_vertex(k_vertex,0,x);
  set_vertex(k_vertex,1,y);
  set_vertex(k_vertex,2,z);
  return 0;
}

int Mesh::set_vertex(int k_vertex,V_3D V)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    set_vertex(k_vertex,k_dim,V[k_dim]);
  return 0;
}

int Mesh::set_vertex_add(int k_vertex,int k_dim,double d_value)
{
  if(3*k_vertex+k_dim<0 || 3*k_vertex+k_dim>=int(vertex.size()))
    {
      printf("Error in set_vertex in Mesh, k_vertex[%d,%d] is not correct for size [%d]\n",k_vertex,k_dim,vertex.size());
      exit(-1);
    }

    vertex[3*k_vertex+k_dim]+=d_value;
    return 0;
}
int Mesh::set_vertex_add(int k_vertex,const V_3D& d_value)
{
  if(3*k_vertex<0 || 3*k_vertex+2>=int(vertex.size()))
    {
      printf("Error in set_vertex in Mesh, k_vertex[%d] is not correct for size [%d]\n",k_vertex,vertex.size());
      exit(-1);
    }
    for(int k_dim=0;k_dim<3;k_dim++)
      vertex[3*k_vertex+k_dim] += d_value[k_dim];
    return 0;
}

int Mesh::get_vertex_size() const
{
  if((int(vertex.size()))%3!=0)
    {printf("something weird in vertex size [%d] in get_vertex_size in Mesh\n",vertex.size());
      exit(-1);
    }
  
  return vertex.size()/3;
}
int Mesh::get_vertex_number() const
{
  return get_vertex_size();
}

double& Mesh::get_vertex(int k_triangle,int k_vertex,int k_dim)
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  if(vertex_number<0 || vertex_number>=int(vertex.size()/3))
    {printf("Error in get_vertex by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return get_vertex(vertex_number,k_dim);
}

double Mesh::get_vertex(int k_triangle,int k_vertex,int k_dim) const
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  if(vertex_number<0 || vertex_number>=int(vertex.size()/3))
    {printf("Error in get_vertex by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return get_vertex(vertex_number,k_dim);
}




V_3D Mesh::get_vertex_v3d(int k_vertex) const
{
  V_3D v(get_vertex(k_vertex,0),get_vertex(k_vertex,1),get_vertex(k_vertex,2));
  return v;
}

V_3D Mesh::get_vertex_v3d(int k_triangle,int k_vertex) const
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  if(vertex_number<0 || 3*vertex_number+2>=int(vertex.size()))
    {printf("Error in get_vertex_v3d by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return get_vertex_v3d(vertex_number);
}



int& Mesh::get_index_vertex(int k_index,int k_dim)
{
  if(3*k_index+k_dim<0 || 3*k_index+k_dim>=int(index_vertex.size()))
    {
      printf("Error in get_index_vertex in Mesh, k_index[%d,%d] is not correct for size [%d]\n",k_index,k_dim,get_triangle_number());
      exit(-1);
    }

    return index_vertex[3*k_index+k_dim];
}

int Mesh::get_index_vertex(int k_index,int k_dim) const
{
  if(3*k_index+k_dim<0 || 3*k_index+k_dim>=int(index_vertex.size()))
    {
      printf("Error in get_index_vertex in Mesh, k_index[%d,%d] is not correct for size [%d]\n",k_index,k_dim,get_triangle_number());
      exit(-1);
    }

    return index_vertex[3*k_index+k_dim];
}


int& Mesh::get_triangle_number(int k_index,int k_dim)
{
  return get_index_vertex(k_index,k_dim);
}

int Mesh::get_triangle_number(int k_index,int k_dim) const
{
  return get_index_vertex(k_index,k_dim);
}



int Mesh::get_index_vertex_size() const
{
  if(int(index_vertex.size())%3!=0)
    {printf("something weird in index_vertex size [%d] in get_vertex_size in Mesh\n",index_vertex.size());
      exit(-1);
    }

  return (index_vertex.size()/3);
}

int Mesh::get_triangle_number() const
{
  return get_index_vertex_size();
}

double& Mesh::get_normal(int k_normal,int k_dim)
{
  if(3*k_normal+k_dim<0 || 3*k_normal+k_dim>=int(normal.size()))
    {
      printf("Error in get_normal in Mesh, k_normal[%d,%d] is not correct for size [%d]\n",k_normal,k_dim,normal.size());
      exit(-1);
    }

  return normal[3*k_normal+k_dim];
}

double Mesh::get_normal(int k_normal,int k_dim) const
{
  if(3*k_normal+k_dim<0 || 3*k_normal+k_dim>=int(normal.size()))
    {
      printf("Error in get_normal in Mesh, k_normal[%d,%d] is not correct for size [%d]\n",k_normal,k_dim,normal.size());
      exit(-1);
    }

  return normal[3*k_normal+k_dim];
}

int Mesh::set_normal(int k_normal,int k_dim,double value)
{
  if(3*k_normal+k_dim<0 || 3*k_normal+k_dim>=int(normal.size()))
    {
      printf("Error in set_normal in Mesh, k_normal[%d,%d] is not correct for size [%d]\n",k_normal,k_dim,normal.size());
      exit(-1);
    }
  normal[3*k_normal+k_dim] = value;
  return 0;
}

int Mesh::set_normal(int k_normal,V_3D V)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    set_normal(k_normal,k_dim,V[k_dim]);
  return 0;
}

double& Mesh::get_normal(int k_triangle,int k_vertex,int k_dim)
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  
  if(vertex_number<0 || 3*vertex_number+2>=int(vertex.size()))
    {printf("Error in get_vertex_v3d by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return normal[3*vertex_number+k_dim];
}

double Mesh::get_normal(int k_triangle,int k_vertex,int k_dim) const
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  
  if(vertex_number<0 || 3*vertex_number+2>=int(vertex.size()))
    {printf("Error in get_vertex_v3d by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return normal[3*vertex_number+k_dim];
}
 
V_3D Mesh::get_normal_v3d(int k_normal) const
{
  V_3D v(get_normal(k_normal,0),get_normal(k_normal,1),get_normal(k_normal,2));
  return v;
}

V_3D Mesh::get_normal_v3d(int k_triangle,int k_vertex) const
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  return get_normal_v3d(vertex_number);
}


double& Mesh::get_color(int k_color,int k_dim)
{
  if(4*k_color+k_dim<0 || 4*k_color+k_dim>=int(color.size()))
    {
      printf("Error in get_color in Mesh, k_color[%d,%d] is not correct for size [%d]\n",k_color,k_dim,color.size());
      exit(-1);
    }

  return color[4*k_color+k_dim];
}

double& Mesh::get_color(int k_triangle,int k_vertex,int k_dim)
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  if(vertex_number<0 || vertex_number>=int(vertex.size()/3))
    {printf("Error in get_vertex by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return get_color(vertex_number,k_dim);
}


double Mesh::get_color(int k_color,int k_dim) const
{
  if(4*k_color+k_dim<0 || 4*k_color+k_dim>=int(color.size()))
    {
      printf("Error in get_color in Mesh, k_color[%d,%d] is not correct for size [%d]\n",k_color,k_dim,color.size());
      exit(-1);
    }

  return color[4*k_color+k_dim];
}

double Mesh::get_color(int k_triangle,int k_vertex,int k_dim) const
{
  int vertex_number = get_index_vertex(k_triangle,k_vertex);
  if(vertex_number<0 || vertex_number>=int(vertex.size()/3))
    {printf("Error in get_vertex by triangle in Mesh, vertex_number[%d] is not correct\n",vertex_number);exit(-1);}

  return get_color(vertex_number,k_dim);
}


int Mesh::set_color(int k_color,int k_dim,double value)
{
  if(4*k_color+k_dim<0 || 4*k_color+k_dim>=int(color.size()))
    {
      printf("Error in set_color in Mesh, k_color[%d,%d] is not correct for size [%d]\n",k_color,k_dim,color.size());
      exit(-1);
    }

  color[4*k_color+k_dim]=value;
  return 0;
}

int Mesh::fill_color(double r,double g,double b,double alpha)
{
  if(int(color.size())!=4*get_vertex_size())
    color.resize(4*get_vertex_size());

  if(color.size()<=0){printf("Warning mesh size is not correct in fill color in Mesh\n\n");return(-1);}
  

  int k_color=0;
  for(k_color=0;k_color<get_vertex_size();k_color++)
    {
      set_color(k_color,0,r);
      set_color(k_color,1,g);
      set_color(k_color,2,b);
      set_color(k_color,3,alpha);
    }

  return 0;

}

int Mesh::fill_color(double r,double g,double b)
{
  if(int(color.size())!=4*get_vertex_size())
    color.resize(4*get_vertex_size());

  if(color.size()<=0){printf("Warning mesh size is not correct in fill color in Mesh\n\n");return(-1);}
  
  int k_color=0;
  for(k_color=0;k_color<get_vertex_size();k_color++)
    {
      set_color(k_color,0,r);
      set_color(k_color,1,g);
      set_color(k_color,2,b);
    }

  return 0;
}

int Mesh::fill_alpha(double alpha)
{
  if(int(color.size())!=4*get_vertex_size())
    color.resize(4*get_vertex_size());

  if(int(color.size())<=0){printf("Warning mesh size is not correct in fill color in Mesh\n\n");return(-1);}
  
  int k_color=0;
  for(k_color=0;k_color<get_vertex_size();k_color++)
    set_color(k_color,3,alpha);

  return 0;
}


int Mesh::build_smooth_normal()
{

  //averaged normal on the one ring

  if(int(normal.size())!=3*get_vertex_size())
    normal.resize(3*get_vertex_size());

  if(normal.size()<=0){printf("Warning mesh size is not correct in build_smooth_normal in Mesh\n\n");return(-1);}
  
  //init
  int k_vertex=0;
  int k_dim=0;
  V_3D n0(0.0,1.0,0.0);
  for(k_vertex=0;k_vertex<get_vertex_size();k_vertex++)
    set_normal(k_vertex,n0);

  int k_triangle=0;
  V_3D v0,v1,v2;
  V_3D n;
  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    {
      v0 = get_vertex_v3d(k_triangle,0);
      v1 = get_vertex_v3d(k_triangle,1);
      v2 = get_vertex_v3d(k_triangle,2);

      n = (v1-v0).vector_prod(v2-v0);
      n = n/n.norm();

      for(k_dim=0;k_dim<3;k_dim++)
	{
	  k_vertex = get_index_vertex(k_triangle,k_dim);
	  set_normal(k_vertex,n+get_normal_v3d(k_vertex));
	}
    }

  //normalize
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    {
      n=get_normal_v3d(k_vertex);
      set_normal(k_vertex,n/n.norm());
    }
  

  return 0;
  
}





int Mesh::add_vertex(double value)
{
  vertex.push_back(value);
  return 0;
}

int Mesh::add_vertex(double x,double y,double z)
{
  vertex.push_back(x);
  vertex.push_back(y);
  vertex.push_back(z);

  return 0;
}

int Mesh::add_vertex(V_3D v)
{
  for(int k_dim=0;k_dim<3;k_dim++)
    vertex.push_back(v[k_dim]);
  return 0;
}

int Mesh::add_triangle(int index_0,int index_1,int index_2)
{
  index_vertex.push_back(index_0);
  index_vertex.push_back(index_1);
  index_vertex.push_back(index_2);

  return 0;
}

int Mesh::add_triangle(std::vector <int> index)
{
  if(index.size()<3 || index.size()%3!=0)
    {printf("Error index size to add in add triangle in Mesh is not correct [%d]\n",index.size());exit(-1);}

  for(unsigned int k_dim=0;k_dim<index.size()/3;k_dim++)
    add_triangle(index[3*k_dim+0],index[3*k_dim+1],index[3*k_dim+2]);

  return 0;
}

int Mesh::add_polygon(const std::vector <int> &index)
{
  if(index.size()<3)
    {printf("Error index size to add in add polygon in Mesh is not correct [%d]\n",index.size());exit(-1);}
  
  int i0=index[0];
  int i1=index[1];
  for(int k_vertex=0;k_vertex<int(index.size())-2;k_vertex++)
    {
      index_vertex.push_back(i0);
      index_vertex.push_back(i1);
      index_vertex.push_back(index[k_vertex+2]);
      
      i1 = index[k_vertex+2];
    }
  

  return 0;
}


int Mesh::fast_check_integrity()
{
  printf("[N_vertex: %d]\n",get_vertex_number());
  printf("[N_triangle: %d]\n",get_triangle_number());

  if(normal.size()<=0)
    printf("[Warning size normal is null [%d]\n",normal.size());
  else if(normal.size()!=vertex.size())
    printf("[Warning size normal is weird [%d]\n",normal.size());

  if(color.size()<=0)
    printf("[Warning size color is null [%d]\n",color.size());
  else if(color.size()/4!=vertex.size()/3)
    printf("[Warning size color is weird [%d]\n",color.size());

  //check index
  int k_triangle=0;
  int k_dim=0;
  int k_vertex=0;
  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    for(k_dim=0;k_dim<3;k_dim++)
      {
	k_vertex = get_triangle_number(k_triangle,k_dim);
	if(k_vertex<0 || k_vertex>=get_vertex_number())
	  {
	    printf("Error in connectivity, k_triangle to large [%d] for [k_triangle=%d,k_dim=%d]\n",k_vertex,k_triangle,k_dim);
	    exit(-1);
	  }
      }
      

  //check normal
  double epsilon=0.00001;
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    if(fabs(get_normal_v3d(k_vertex).norm()-1.0) > epsilon)
      {
	printf("Error no unitary normal [%f] for [k_normal=%d]\n",get_normal_v3d(k_vertex).norm(),k_vertex);
	exit(-1);
      }

  printf("[Fast check integrity OK]\n");

  return 0;
}


Mesh& Mesh::operator=(const Mesh& _M)
{
  destroy();

  unsigned int k=0;

  vertex.resize(_M.vertex.size());
  for(k=0;k<vertex.size();k++)
    vertex[k]=_M.vertex[k];

  index_vertex.resize(_M.index_vertex.size());
  for(k=0;k<index_vertex.size();k++)
    index_vertex[k]=_M.index_vertex[k];

  normal.resize(_M.normal.size());
  for(k=0;k<normal.size();k++)
    normal[k]=_M.normal[k];

  color.resize(_M.color.size());
  for(k=0;k<color.size();k++)
    color[k]=_M.color[k];

  one_ring.resize(_M.one_ring.size());
  for(k=0;k<one_ring.size();k++)
    one_ring[k]=_M.one_ring[k];

  two_ring.resize(_M.two_ring.size());
  for(k=0;k<two_ring.size();k++)
    two_ring[k]=_M.two_ring[k];


  return *this;
}


int Mesh::unitize_size()
{
  V_3D mini(999,999,999),maxi(-999,-999,-999);
  V_3D V;
  V_3D L;
  
  int k_vertex=0;
  int k_dim=0;
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    {
      V = get_vertex_v3d(k_vertex);
      for(k_dim=0;k_dim<3;k_dim++)
	{
	  if(mini[k_dim]>V[k_dim])
	    mini[k_dim] = V[k_dim];
	  if(maxi[k_dim]<V[k_dim])
	    maxi[k_dim] = V[k_dim];
	}
    }
  L = maxi-mini;
  double L_max=0.0;
  for(k_dim=0;k_dim<3;k_dim++)
    if(L_max<L[k_dim])
      L_max=L[k_dim];

  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    for(k_dim=0;k_dim<3;k_dim++)
      get_vertex(k_vertex,k_dim) = ((get_vertex(k_vertex,k_dim)-mini[k_dim])/L_max-L[k_dim]/(2*L_max))*2;

  
  return 0;
}

int Mesh::reverse_triangle_order()
{
  int k_triangle=0;
  int temp;
  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    {
      temp = get_index_vertex(k_triangle,1);
      get_index_vertex(k_triangle,1)=get_index_vertex(k_triangle,2);
      get_index_vertex(k_triangle,2)=temp;
    }

  return 0;
}


int Mesh::build_fast_visualization_pointers()
{

#ifdef DEBUG
  printf("\n[Build Fast Visualization Pointers]\n\n");
#endif 

  empty_pointers();
  is_fast_visualization_pointers=1;

  p_vertex=NULL;
  p_normal=NULL;
  p_color=NULL;
  p_index_vertex=NULL;

  fast_check_integrity();

  p_vertex = new double[vertex.size()];
  p_normal = new double[normal.size()];
  p_color = new double[color.size()];
  p_index_vertex = new int[index_vertex.size()];

  if(p_vertex==NULL){printf("Error allocating p_vertex in build_fast_visualization_pointers in Mesh\n");exit(-1);}
  if(p_normal==NULL){printf("Error allocating p_normal in build_fast_visualization_pointers in Mesh\n");exit(-1);}
  if(p_color==NULL){printf("Error allocating p_color in build_fast_visualization_pointers in Mesh\n");exit(-1);}
  if(p_index_vertex==NULL){printf("Error allocating p_index_vertex in build_fast_visualization_pointers in Mesh\n");exit(-1);}

  unsigned int k=0;
  for(k=0;k<vertex.size();k++)
    p_vertex[k] = vertex[k];
  for(k=0;k<normal.size();k++)
    p_normal[k] = normal[k];  
  for(k=0;k<color.size();k++)
    p_color[k] = color[k];
  for(k=0;k<index_vertex.size();k++)
    p_index_vertex[k] = index_vertex[k];

  return 0;
}

int Mesh::empty_pointers()
{
  is_fast_visualization_pointers=0;
  if(p_vertex!=NULL)
    {
      delete []p_vertex;
      p_vertex=NULL;
    }
  if(p_normal!=NULL)
    {
      delete []p_normal;
      p_normal=NULL;
    }
  if(p_color!=NULL)
    {
      delete []p_color;
      p_color=NULL;
    }
  if(p_index_vertex!=NULL)
    {						
      delete []p_index_vertex;
      p_index_vertex=NULL;
    }
return 0;
}

int Mesh::is_builded_pointers()
{
  return is_fast_visualization_pointers;
}

double *Mesh::get_p_vertex()
{
  if(p_vertex==NULL)
    {printf("Error p_vertex is null in get_p_vertex in Mesh\n");exit(-1);}
  return p_vertex;
}

double *Mesh::get_p_normal()
{
  if(p_normal==NULL)
    {printf("Error p_normal is null in get_p_normal in Mesh\n");exit(-1);}
  return p_normal;
}

double *Mesh::get_p_color()
{
  if(p_color==NULL)
    {printf("Error p_color is null in get_p_color in Mesh\n");exit(-1);}
  return p_color;
}

int *Mesh::get_p_index_vertex()
{
  if(p_index_vertex==NULL)
    {printf("Error p_index_vertex is null in get_p_index_vertex in Mesh\n");exit(-1);}
  return p_index_vertex;
}

int Mesh::is_color() const
{
  if(3*color.size()!=4*vertex.size())
    return 0;
  else
    return 1;
}

V_3D Mesh::get_min_point() const
{
  V_3D min_point(9999,9999,9999);
  V_3D current;
  int k_vertex=0,N_vertex=get_vertex_number();
  int k_dim=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      current = get_vertex_v3d(k_vertex);
      for(k_dim=0;k_dim<3;k_dim++)
	if(current[k_dim]<min_point[k_dim])
	  min_point[k_dim] = current[k_dim];
    }

  return min_point;

}

V_3D Mesh::get_max_point() const
{
  V_3D max_point(-9999,-9999,-9999);
  V_3D current;
  int k_vertex=0,N_vertex=get_vertex_number();
  int k_dim=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      current = get_vertex_v3d(k_vertex);
      for(k_dim=0;k_dim<3;k_dim++)
	if(current[k_dim]>max_point[k_dim])
	  max_point[k_dim] = current[k_dim];
    }
  return max_point;
}

V_3D Mesh::get_middle_point() const
{
  V_3D min_point = get_min_point();
  V_3D max_point = get_max_point();

  return 0.5*(min_point+max_point);  
}

V_3D Mesh::get_size_mesh() const
{
  V_3D min_point = get_min_point();
  V_3D max_point = get_max_point();

  return (max_point-min_point); 
}

int Mesh::rotate(const V_3D& n,double angle)
{return rotate(n[0],n[1],n[2],angle);}
int Mesh::rotate(double *n,double angle)
{return rotate(n[0],n[1],n[2],angle);}
int Mesh::rotate(double n_x,double n_y,double n_z,double angle)
{
  Matrix R;
  R.set_rotation_matrix(n_x,n_y,n_z,angle);

  int N_vertex=get_vertex_number();
  int k_vertex=0;

  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    set_vertex(k_vertex,R*get_vertex_v3d(k_vertex));
  return 0;
}
int Mesh::translate(double t_x,double t_y,double t_z)
{
  V_3D T(t_x,t_y,t_z);
  int N_vertex=get_vertex_number();
  int k_vertex=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    set_vertex(k_vertex,get_vertex_v3d(k_vertex)+T);
  return 0;
}
int Mesh::translate(double *t)
{return translate(t[0],t[1],t[2]);}
int Mesh::scale(double s_x,double s_y,double s_z)
{
  int N_vertex=get_vertex_number();
  int k_vertex=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      set_vertex(k_vertex,0,get_vertex(k_vertex,0)*s_x);
      set_vertex(k_vertex,1,get_vertex(k_vertex,0)*s_y);
      set_vertex(k_vertex,2,get_vertex(k_vertex,0)*s_z);
    }
  return 0;
}
int Mesh::scale(double s)
{
  int N_vertex=get_vertex_number();
  int k_vertex=0;
  int k_dim=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    for(k_dim=0;k_dim<3;k_dim++)
      set_vertex(k_vertex,k_dim,get_vertex(k_vertex,k_dim)*s);
  return 0;
}
int Mesh::scale(double *s)
{return scale(s[0],s[1],s[2]);}


int Mesh::build_one_ring()
{

  one_ring.resize(0);
  one_ring.resize(get_vertex_number());
  
  int k_triangle=0;
  int N_triangle=get_triangle_number();
  int index[3]={0,0,0};
  int k_dim=0;
  for(k_triangle=0;k_triangle<N_triangle;k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	index[k_dim] = get_index_vertex(k_triangle,k_dim);

      one_ring[index[0]].add_unique(index[1]);
      one_ring[index[0]].add_unique(index[2]);

      one_ring[index[1]].add_unique(index[0]);
      one_ring[index[1]].add_unique(index[2]);

      one_ring[index[2]].add_unique(index[0]);
      one_ring[index[2]].add_unique(index[1]);
    }

  //build two_ring
  int N_vertex=get_vertex_number();
  int k_neighbors_1=0,N_neighbors_1=0;
  int k_neighbors_2=0,N_neighbors_2=0;
  int k_vertex=0,k_vertex_2=0;
  two_ring.resize(0);two_ring.resize(N_vertex);
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      N_neighbors_1=get_neighbors_number(k_vertex);
      for(k_neighbors_1=0;k_neighbors_1<N_neighbors_1;k_neighbors_1++)
	{
	  k_vertex_2=get_neighbors(k_vertex,k_neighbors_1);
	  N_neighbors_2=get_neighbors_number(k_vertex_2);
	  for(k_neighbors_2=0;k_neighbors_2<N_neighbors_2;k_neighbors_2++)
	    if(get_neighbors(k_vertex_2,k_neighbors_2)!=k_vertex)
	      two_ring[k_vertex].add_unique(get_neighbors(k_vertex_2,k_neighbors_2));
	}
      
    }


  return 0;
}

int Mesh::get_neighbors_number(int k_vertex) const
{
  if(k_vertex<0 || k_vertex>get_vertex_number())
    {printf("Error k_vertex [%d] is not correct in get_neighbors_number in Mesh\n",k_vertex);exit(-1);}
  return one_ring[k_vertex].size();
}
int Mesh::get_neighbors(int k_vertex,int k_neighbors) const
{
  if(k_vertex<0 || k_vertex>get_vertex_number())
    {printf("Error k_vertex [%d] is not correct in get_neighbors in Mesh\n",k_vertex);exit(-1);}
  return one_ring[k_vertex][k_neighbors];

}


int Mesh::get_neighbors_two_ring_number(int k_vertex) const
{
 if(k_vertex<0 || k_vertex>get_vertex_number())
    {printf("Error k_vertex [%d] is not correct in get_neighbors_two_ring_number in Mesh\n",k_vertex);exit(-1);}
 return two_ring[k_vertex].size();
}
int Mesh::get_neighbors_two_ring(int k_vertex,int k_neighbors) const
{
  if(k_vertex<0 || k_vertex>get_vertex_number())
    {printf("Error k_vertex [%d] is not correct in get_neighbors_two_ring in Mesh\n",k_vertex);exit(-1);}
  return two_ring[k_vertex][k_neighbors];
}

int Mesh::set_neighbors(int k_vertex,int k_neighbors,int index)
{
  if(k_vertex<0 || k_vertex>get_vertex_number())
    {printf("Error k_vertex [%d] is not correct in set_neighbors in Mesh\n",k_vertex);exit(-1);}  
  one_ring[k_vertex][k_neighbors]=index;
  return 0;
}




int Mesh::smooth_surface()
{

  if(int(one_ring.size())!=get_vertex_number())
    build_one_ring();


  int k_vertex=0;
  V_3D bar;
  int k_neighbors=0;
  int N_vertex=get_vertex_number();
  int N_neighbors=0;
  double omega=0.0;
  double lambda=0.2;


  std::vector <V_3D> temp;
  temp.resize(N_vertex);

  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      bar.set(0.0,0.0,0.0);
      N_neighbors=one_ring[k_vertex].size();
      omega=1.0/double(N_neighbors);
      for(k_neighbors=0;k_neighbors<N_neighbors;k_neighbors++)
	bar += omega*get_vertex_v3d(one_ring[k_vertex][k_neighbors]);
      temp[k_vertex]=get_vertex_v3d(k_vertex)+lambda*(bar-get_vertex_v3d(k_vertex));
    }
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    set_vertex(k_vertex,temp[k_vertex]);


  return 0;
}


int Mesh::subdivide_surface()
{
  Mesh mesh2;
  int k_triangle=0;
  int N_triangle = get_triangle_number();
  int N_vertex = get_vertex_number();
  int index[3]={0,0,0};
  int k_dim=0;
  int count=0;
  std::vector <int> linked_index;
  int new_index[3]={0,0,0};
  linked_index.resize(0);
  for(k_triangle=0;k_triangle<N_triangle;k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	index[k_dim]=get_triangle_number(k_triangle,k_dim);

      new_index[0] = mesh2.create_linked_index_if_needed(index[0],index[1],&linked_index,&count);
      new_index[1] = mesh2.create_linked_index_if_needed(index[1],index[2],&linked_index,&count);
      new_index[2] = mesh2.create_linked_index_if_needed(index[2],index[0],&linked_index,&count);


      //add new triangles //split each triangle into 4 news
      mesh2.add_triangle(index[0],N_vertex+new_index[0],N_vertex+new_index[2]);
      mesh2.add_triangle(N_vertex+new_index[0],index[1],N_vertex+new_index[1]);
      mesh2.add_triangle(N_vertex+new_index[1],index[2],N_vertex+new_index[2]);
      mesh2.add_triangle(N_vertex+new_index[0],N_vertex+new_index[1],N_vertex+new_index[2]);

    }


  int k_vertex=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    mesh2.add_vertex(get_vertex_v3d(k_vertex));
  

  laplace_scheme(&mesh2,linked_index,count);


  *this=mesh2;
  build_smooth_normal();

  return 0;
}

int Mesh::laplace_scheme(Mesh *mesh2,const std::vector <int>& linked_index,int N_new_vertex)
{
  int k_vertex=0;

  for(k_vertex=0;k_vertex<N_new_vertex;k_vertex++)
    mesh2->add_vertex(0.5*(get_vertex_v3d(linked_index[2*k_vertex+0])+get_vertex_v3d(linked_index[2*k_vertex+1])));
  for(int k=0;k<10;k++)
    mesh2->smooth_surface();

  return 0;
}

int Mesh::create_linked_index_if_needed(int index_0,int index_1,std::vector <int> *linked_index,int *count)
{
  int k_linked_index=0;
  int N_linked_index = int(linked_index->size()/2.0);

  for(k_linked_index=0;k_linked_index<N_linked_index;k_linked_index++)
    {
      if(
	 ((*linked_index)[2*k_linked_index+0]==index_0 && (*linked_index)[2*k_linked_index+1]==index_1)
	 ||
	 ((*linked_index)[2*k_linked_index+0]==index_1 && (*linked_index)[2*k_linked_index+1]==index_0)
	 )
	return k_linked_index;//if we found the linked index
    }
  
  //if not, create a new one
  linked_index->push_back(index_0);
  linked_index->push_back(index_1);
  *count=*count+1;

  return k_linked_index;
}

int Mesh::cut_plane(const V_3D& n,const V_3D& x0)
{return cut_plane(n[0],n[1],n[2],-n.dot(x0));}
int Mesh::cut_plane(double a,double b,double c,double d)
{V_3D x(a,b,c);return cut_plane(x,d);}

int Mesh::cut_plane(const V_3D& n,double d)
{
  Mesh mesh2;

  int k_vertex=0;
  std::vector <int> is_cutted;
  double value=0.0;
  V_3D x;

  is_cutted.resize(get_vertex_number());

  //mark every vertices which are cutted
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    {
      is_cutted[k_vertex]=0;
      x = get_vertex_v3d(k_vertex);
      value = n.dot(x)+d;
      value>=0 ? is_cutted[k_vertex]=0 : is_cutted[k_vertex]=1;

      mesh2.add_vertex(x);
    }

  //put away the cutted triangles
  int t[3]={0,0,0};
  int k_dim=0,k_triangle=0;
  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	t[k_dim] = get_triangle_number(k_triangle,k_dim);

      //add triangle if every vertices are not cutted
      if(is_cutted[t[0]]==0 && is_cutted[t[1]]==0 && is_cutted[t[2]]==0)
	mesh2.add_triangle(t[0],t[1],t[2]);
    }

  mesh2.build_smooth_normal();
  *this=mesh2;

  return 0;
}


int Mesh::select_box(const V_3D& p1,const V_3D& p2)
{
  Mesh mesh2;

  int k_vertex=0;
  std::vector <int> is_cutted;
  V_3D x;

  is_cutted.resize(get_vertex_number());


  V_3D p_min = p1.get_min_scalar(p2);
  V_3D p_max = p1.get_max_scalar(p2);

  //mark every vertices which are cutted
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    {
      is_cutted[k_vertex]=0;
      x = get_vertex_v3d(k_vertex);

      if(x[0]>=p_min[0] && x[0]<=p_max[0] 
	 && x[1]>=p_min[1] && x[1]<=p_max[1]
	 && x[2]>=p_min[2] && x[2]<=p_max[2])
	{
	  is_cutted[k_vertex]=0;
	}
      else
	is_cutted[k_vertex]=1;

      mesh2.add_vertex(x);
    }

  //put away the cutted triangles
  int t[3]={0,0,0};
  int k_dim=0,k_triangle=0;
  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	t[k_dim] = get_triangle_number(k_triangle,k_dim);

      //add triangle if every vertices are not cutted
      if(is_cutted[t[0]]==0 && is_cutted[t[1]]==0 && is_cutted[t[2]]==0)
	mesh2.add_triangle(t[0],t[1],t[2]);
    }

  mesh2.build_smooth_normal();
  *this=mesh2;

  return 0;
}


int Mesh::keep_triangles(int tri_start,int tri_end)
{
  Mesh mesh2;
  int k_vertex=0,k_triangle=0,k_dim=0;
  int t[3]={0,0,0};
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    mesh2.add_vertex(get_vertex_v3d(k_vertex));

  //put away the triangles eliminated
  for(k_triangle=tri_start;k_triangle<tri_end;k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	t[k_dim] = get_triangle_number(k_triangle,k_dim);
      mesh2.add_triangle(t[0],t[1],t[2]);
    }

  mesh2.build_smooth_normal();
  *this=mesh2;

  return 0;
}

int Mesh::delete_triangles(int tri_start,int tri_end)
{
  Mesh mesh2;
  int k_vertex=0,k_triangle=0,k_dim=0;
  int t[3]={0,0,0};
  for(k_vertex=0;k_vertex<get_vertex_number();k_vertex++)
    mesh2.add_vertex(get_vertex_v3d(k_vertex));


  for(k_triangle=0;k_triangle<get_triangle_number();k_triangle++)
    {
      if(k_triangle<tri_start || k_triangle>tri_end)
	{
	  for(k_dim=0;k_dim<3;k_dim++)
	    t[k_dim] = get_triangle_number(k_triangle,k_dim);
	  mesh2.add_triangle(t[0],t[1],t[2]);
	}
    }

  mesh2.build_smooth_normal();
  *this=mesh2;

  return 0;
}


int Mesh::load_connectivity(const Mesh& _mesh)
{
  destroy_connectivity();
  int k_triangle=0,N_triangle=_mesh.get_triangle_number();
  for(k_triangle=0;k_triangle<N_triangle;k_triangle++)
    add_triangle(_mesh.get_triangle_number(k_triangle,0),_mesh.get_triangle_number(k_triangle,1),_mesh.get_triangle_number(k_triangle,2));

  return 0;
}


double Mesh::norm_2(const Mesh& _mesh) const
{
  if(get_vertex_number()!=_mesh.get_vertex_number())
    {printf("Error in norm_2 in Mesh, size are not compatible [%d]-[%d]\n",get_vertex_number(),_mesh.get_vertex_number());exit(-1);}

  double norm=0.0;
  V_3D p1;
  V_3D p2;
  int k_vertex=0,N_vertex=get_vertex_number();
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      p1=get_vertex_v3d(k_vertex);
      p2=_mesh.get_vertex_v3d(k_vertex);
      
      norm += (p1-p2).dot(p1-p2);
    }
  powf(norm,0.5);


  return norm;
}


//build circle and the boundary is returned
std::vector<int> Mesh::build_circle(int N_decompose)
{


  destroy();
  
//   V_3D x1(0.0,1.0,0.0);
//   V_3D x2(-powf(3.0,0.5)/2.0,-1.0/2.0,0.0);
//   V_3D x3( powf(3.0,0.5)/2.0,-1.0/2.0,0.0);
//   add_vertex(x1);add_vertex(x2);add_vertex(x3);
//   add_triangle(0,1,2);

  double p0=1.0/powf(2.0,0.5);
  V_3D x0(0.0,0.0,0.0);
  V_3D x1(-p0,-p0,0.0);
  V_3D x2( p0,-p0,0.0);
  V_3D x3( p0, p0,0.0);
  V_3D x4(-p0, p0,0.0);
  add_vertex(x0);add_vertex(x1);add_vertex(x2);add_vertex(x3);add_vertex(x4);
  add_triangle(1,2,0);add_triangle(2,3,0);add_triangle(3,4,0);add_triangle(4,1,0);


  std::list <int> boundary;
  //boundary.push_back(0);boundary.push_back(1);boundary.push_back(2);
  boundary.push_back(1);boundary.push_back(2);boundary.push_back(3);boundary.push_back(4);

  Mesh mesh2;

  int k_decompose=0;
  for(k_decompose=0;k_decompose<N_decompose;k_decompose++)
    {
      mesh2.destroy();
      
      //subdivide
      std::vector <int> linked_index;
      int new_index[3]={0,0,0};
      linked_index.resize(0);
      int k_triangle=0,N_triangle=get_triangle_number(),N_vertex=get_vertex_number();
      int k_dim=0,index[3]={0,0,0};
      int count=0;
      for(k_triangle=0;k_triangle<N_triangle;k_triangle++)
	{
	  for(k_dim=0;k_dim<3;k_dim++)
	    index[k_dim]=get_triangle_number(k_triangle,k_dim);
	  
	  new_index[0] = mesh2.create_linked_index_if_needed(index[0],index[1],&linked_index,&count);
	  new_index[1] = mesh2.create_linked_index_if_needed(index[1],index[2],&linked_index,&count);
	  new_index[2] = mesh2.create_linked_index_if_needed(index[2],index[0],&linked_index,&count);
	  
	  
	  //add new triangles //split each triangle into 4 news
	  mesh2.add_triangle(index[0],N_vertex+new_index[0],N_vertex+new_index[2]);
	  mesh2.add_triangle(N_vertex+new_index[0],index[1],N_vertex+new_index[1]);
	  mesh2.add_triangle(N_vertex+new_index[1],index[2],N_vertex+new_index[2]);
	  mesh2.add_triangle(N_vertex+new_index[0],N_vertex+new_index[1],N_vertex+new_index[2]);
	  
	}
      
      
      int N_new_index = linked_index.size()/2;
      int k_vertex=0;
      for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
	mesh2.add_vertex(get_vertex_v3d(k_vertex));

      //boundary
      int current_index=0;
      for(k_vertex=0;k_vertex<N_new_index;k_vertex++)
	{
	  current_index=N_vertex+k_vertex;
	  build_boundary_edge(&boundary,linked_index[2*k_vertex+0],linked_index[2*k_vertex+1],current_index);
	}

      
      //split half size
      for(k_vertex=0;k_vertex<N_new_index;k_vertex++)
	{
	  mesh2.add_vertex(0.5*(get_vertex_v3d(linked_index[2*k_vertex+0])+get_vertex_v3d(linked_index[2*k_vertex+1])));
	  if(is_boundary(boundary,k_vertex+N_vertex)==1)
	    {
	      V_3D v=mesh2.get_vertex_v3d(k_vertex+N_vertex);
	      v = v/v.norm();
	      mesh2.set_vertex(k_vertex+N_vertex,v);
	    }
	}      
      
      
      *this=mesh2;
    }
  

  
  std::vector <int> v_boundary;
  std::list<int>::iterator it=boundary.begin();
  for(it=boundary.begin();it!=boundary.end();++it)
    v_boundary.push_back(*it);




  return v_boundary;
}

int Mesh::build_boundary_edge(std::list <int>* boundary,int index_0,int index_1,int current_index)
{
  std::list<int>::const_iterator it_0=boundary->begin();
  std::list<int>::iterator it_1=boundary->begin();++it_1;
  int k=0,N_size=boundary->size();
  for(k=0;k<N_size-1;k++,++it_0,++it_1)
    if((*it_0==index_0 && *it_1==index_1) || (*it_0==index_1 && *it_1==index_0))
      {
	boundary->insert(it_1,current_index);
	return 1;
      }
  //last case
  if((*it_0==index_0 && *(boundary->begin())==index_1) || (*it_0==index_1 && *(boundary->begin())==index_0))
    {
      boundary->push_back(current_index);
      return 1;
    }

  return 0;
}

int Mesh::is_boundary(const std::list <int>& boundary,int k_index)
{
  std::list<int>::const_iterator it=boundary.begin();
  for(it=boundary.begin();it!=boundary.end();++it)
    if(*it==k_index)
      return 1;
  return 0;
}




int Mesh::is_neighbors(int k_vertex,int index)
{
  if(k_vertex<0 || k_vertex>int(one_ring.size()))
    {printf("Error k_vertex [%d] is not correct in is_neighbors (one_ring) in Mesh\n",k_vertex);exit(-1);}

  int k_neighbors=0,N_neighbors=get_neighbors_number(k_vertex);
  for(k_neighbors=0;k_neighbors<N_neighbors;k_neighbors++)
    if(index==get_neighbors(k_vertex,k_neighbors))
      return 1;
  return 0;
}

int Mesh::is_neighbors_two_ring(int k_vertex,int index)
{
  if(k_vertex<0 || k_vertex>int(two_ring.size()))
    {printf("Error k_vertex [%d] is not correct in is_neighbors_two_ring in Mesh\n",k_vertex);exit(-1);}

  int k_neighbors=0,N_neighbors=get_neighbors_two_ring_number(k_vertex);
  for(k_neighbors=0;k_neighbors<N_neighbors;k_neighbors++)
    if(index==get_neighbors_two_ring(k_vertex,k_neighbors))
      return 1;
  return 0;
}

std::vector <int_vector> Mesh::get_one_ring()
{
  if(int(one_ring.size())!=get_vertex_number())
    build_one_ring();
  return one_ring;
}
std::vector <int_vector> Mesh::get_two_ring()
{
  if(int(two_ring.size())!=get_vertex_number())
    build_one_ring();

  return two_ring;
}

std::vector<int> Mesh::build_square(int N_1,int N_2)
{
  destroy();
  
  
  int k_1=0,k_2=0;
  V_3D x;
  for(k_1=0;k_1<N_2;k_1++)
    for(k_2=0;k_2<N_2;k_2++)
      {
	x.set(k_1/double(N_1-1),k_2/double(N_2-1),0.0);
	add_vertex(x);
      }

  for(k_1=0;k_1<N_2-1;k_1++)
    for(k_2=0;k_2<N_2-1;k_2++)
      {
	add_triangle(k_2+N_2*k_1,(k_2+1)+N_2*k_1,(k_2+1)+N_2*(k_1+1));
	add_triangle(k_2+N_2*k_1,(k_2+1)+N_2*(k_1+1),k_2+N_2*(k_1+1));
      }  

  std::vector <int> boundary;
  for(k_1=0;k_1<N_1;k_1++)
    boundary.push_back(0+N_2*k_1);
  for(k_2=0;k_2<N_2;k_2++)
    boundary.push_back(k_2+N_2*(N_1-1));
  for(k_1=N_1-1;k_1>=0;k_1--)
    boundary.push_back((N_2-1)+N_2*k_1);
  for(k_2=N_2-1;k_2>=0;k_2--)
    boundary.push_back(k_2+0);
  
  return boundary;
}


int Mesh::get_closest_vertex(double x,double y,double z) 
{
  V_3D X(x,y,z);
  V_3D v;
  int k_vertex=0;
  double distance=0;
  double min_distance=9999.0;
  int k_min_index=0;
  int N_vertex=get_vertex_number();
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      v = get_vertex_v3d(k_vertex);
      distance = (X-v).norm();
      if(min_distance>distance)
	{
	  min_distance=distance;
	  k_min_index = k_vertex;
	}

    }
  return k_min_index;
}


int Mesh::increase_alpha(double d_alpha)
{
  int k_color=0;
  int N_color=color.size();
  for(k_color=0;k_color<N_color;k_color++)
    {
      color[4*k_color+3]+= d_alpha;
      color[4*k_color+3] = max(min(color[4*k_color+3],1.0),0.0);
    }
  return 0;
}

const double* Mesh::get_vertex()
{
  return &vertex[0];
}
const double* Mesh::get_normal()
{
  return &normal[0];
}
const double* Mesh::get_color()
{
  return &color[0];
}
const int* Mesh::get_index_vertex()
{
  return &index_vertex[0];
}

int Mesh::build_adjacent_faces()
{
  int N_faces = get_index_vertex_size();
  int k_faces;
  adjacent_faces.resize(3*N_faces);
  
  for(k_faces=0;k_faces<N_faces;k_faces++)
    {
      adjacent_faces[3*k_faces+0] = find_opposite_face(k_faces,0,1);
      adjacent_faces[3*k_faces+1] = find_opposite_face(k_faces,1,2);
      adjacent_faces[3*k_faces+2] = find_opposite_face(k_faces,2,0);
    }

  return 0;
}

int Mesh::find_opposite_face(int current_face,int edge_start,int edge_end) const
{
  int N_face=get_index_vertex_size();
  int k_face=0;

  int vertex_to_find[2]={get_index_vertex(current_face,edge_start),get_index_vertex(current_face,edge_end)};
  int current_vertex[3];

  for(k_face=0;k_face<N_face;k_face++)
    if(k_face!=current_face)
      {
	current_vertex[0]=get_index_vertex(k_face,0);
	current_vertex[1]=get_index_vertex(k_face,1);
	current_vertex[2]=get_index_vertex(k_face,2);

	if(
	   (vertex_to_find[0]==current_vertex[0] && vertex_to_find[1]==current_vertex[2])
	   ||
	   (vertex_to_find[0]==current_vertex[1] && vertex_to_find[1]==current_vertex[0])
	   ||
	   (vertex_to_find[0]==current_vertex[2] && vertex_to_find[1]==current_vertex[1])
	   )
	  return k_face;
      }


  return -1;
}

int Mesh::get_adjacent_faces(int k_face,int k_edge) const
{
  if(k_face<0 || k_face>=get_index_vertex_size() || k_edge<0 || k_edge>2)
    {printf("Error k_face/k_edge too large [%d,%d] (size (%d,%d)) in get_adjacent_faces in Mesh\n",k_face,k_edge,get_index_vertex_size(),3);exit(-1);}
  if(k_face>=int(adjacent_faces.size()))
    {printf("Warning adjacent face have not been calculated yet\n");exit(-1);}

  return adjacent_faces[3*k_face+k_edge];
}

std::vector <int> Mesh::get_connectivity_vrml()
{
  std::vector <int> connectivity_vrml;
  int N_triangle=get_triangle_number();
  int k_triangle=0;
  int k_dim=0;

  //index = [vertex1,vertex2,vertex3,-1]
  for(k_triangle=0;k_triangle<N_triangle;k_triangle++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	connectivity_vrml.push_back(get_triangle_number(k_triangle,k_dim));
      connectivity_vrml.push_back(-1);
    }



  return connectivity_vrml;
}
