/** Main program
*/
#include <Inventor/Qt/SoQt.h>
#include <Inventor/Qt/viewers/SoQtExaminerViewer.h>

#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#include <string>
using std::string;
#include <stdlib.h>
#include "argumentParser.h"


#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodekits/SoShapeKit.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoLocation2Event.h>
#include <Inventor/sensors/SoTimerSensor.h>

typedef SoQtExaminerViewer ExaminerViewer;
         ExaminerViewer *myViewer;
         
#ifdef WIN32
    HWND myWindow;
#else
    QWidget * myWindow;
#endif
    
#include "Animation.h"
Animation* animation;
         SoTimerSensor* animClock;
         float t=0;
         float dt = 0.01;
         bool button1=false;
         bool button2=false;
         bool button3=false;



SoSeparator* root=0;
SoEventCallback* eventNode=0;
SoSeparator* dataRoot=0;


// event callbacks 
void myKeypressCallback(void *, SoEventCallback *eventCB);
void myMouseButtonCallback(void *, SoEventCallback *eventCB);
void myMouseMoveCallback(void *, SoEventCallback *eventCB);



/// Method used to handle keyboard events
void myKeypressCallback(void *, SoEventCallback *eventCB)
{
    const SoEvent *event = eventCB->getEvent();
    assert( event->isOfType(SoKeyboardEvent::getClassTypeId() ) );
    const SoKeyboardEvent* kbev = (SoKeyboardEvent*)( event );
    assert( kbev );

    //cerr<<"event key " << kbev->getKey() << endl;

    if( SoKeyboardEvent::isKeyPressEvent(event,SoKeyboardEvent::W ) )
    {
/*        cerr<<"save scene in file out.iv" << endl;
        scene->writeFile("out.iv");*/
        const SbViewportRegion &vp  = myViewer->getViewportRegion();
        SoOffscreenRenderer* renderer = new SoOffscreenRenderer(vp);
        SoNode * theRoot = myViewer->getSceneManager()->getSceneGraph();
        if( ! renderer->render(theRoot) ){
            cerr<<"Unable to render image"<<endl;
        }
        else {
/*            cerr<<"rendered image"<<endl;
            renderer->writeToRGB("capture.rbg");*/
        }
        delete renderer;
        eventCB->setHandled();
    }
    else if( SoKeyboardEvent::isKeyPressEvent(event,SoKeyboardEvent::I ) )
    {
        eventCB->setHandled();
    }

}




/// Method used to handle mouse button events
void myMouseButtonCallback(void */*userData*/, SoEventCallback *eventCB)
{
    const SoEvent *event = eventCB->getEvent();
    static bool running = false;

    if ( SoMouseButtonEvent::isButtonPressEvent(event,SoMouseButtonEvent:: BUTTON1) )
    {
        button1 = true;
        running = !running;
        if( running )
            animClock->schedule();
        else
            animClock->unschedule();
        eventCB->setHandled();
    }
    else if ( SoMouseButtonEvent::isButtonReleaseEvent(event,SoMouseButtonEvent:: BUTTON2) )
    {
        eventCB->setHandled();
    }

}

// Method used to handle mouse motion events
void myMouseMoveCallback(void *userData, SoEventCallback* eventCB )
{

  const SoEvent *event;
  event= eventCB->getEvent();

    ExaminerViewer* v = (ExaminerViewer*) userData;
    if( v->getCamera()->isOfType(SoPerspectiveCamera::getClassTypeId() ) )
    {
      //SbVec2s location = event->getPosition();
	//printf("%f,%f,%f\n",v->getCamera()->position.getValue()[0],v->getCamera()->position.getValue()[1],v->getCamera()->position.getValue()[2]);

        //std::cout<<"mouse now at "<<location[0]<<", "<<location[1]<<endl;
//        scene->manipulate(location);
    }
    else
        cerr<<"can not handle this type of camera"<<endl;
}

// update the gui
void
processGUI(void *, SoSensor *sensor)
{
#ifdef WIN32
    QCoreApplication::processEvents();
#else
    //SoQt::processEvents();
#endif

    sensor->schedule(); 
}


// scene update triggerd by the  SoTimerSensor* animClock
void update(void *, SoSensor *)
{
    t += dt;
    animation->update(t,dt);
}


int main(int argc, char *argv[])
{
    std::string sceneFileName;

    //  ========= Parse the command line
    oi::parse("This program implements a physical simulation.The default scene represents a fountain.\nUser interaction:\n left click: start/stop\n Space: one time step forwards\n W: write scene to file out.iv\n")
    .option(&sceneFileName,'f',"fileName","scene file to visualize. If not scene, build the default scene")
    (argc,argv);

    //  =========  Initialize Inventor and Window manager
#ifdef WIN32
    myWindow	= SoWin::init(argv[0]); // use SoWin 
    if (myWindow == NULL)
        exit(1);
    //Q_INIT_RESOURCE(simpletreemodel);
    QApplication app(argc, argv);
#else
    myWindow = SoQt::init(argc, argv, argv[0]); // use SoQt 
    if (myWindow == NULL)
        exit(1);
#endif
    SoInput::addDirectoryFirst("models");

    // =========  Create the scenegraph

    root = new SoSeparator;
    root->ref();
    
    //  ========= Event capture
    //- callback dispatcher 
    eventNode = new SoEventCallback;
    eventNode->setName("eventNode");
    root->addChild( eventNode );
    // - keyboard callback 
    eventNode->addEventCallback(SoKeyboardEvent::getClassTypeId(), myKeypressCallback, NULL); 



    // =========  build the data
        dataRoot = new SoSeparator;  

    animation = new Animation( dataRoot );
    //dataRoot->addChild( dolphin->getRoot() );    
    
    root->addChild( dataRoot );



    // simulated time
    animClock = new SoTimerSensor( update, NULL ); // callback d'animation
    animClock->setInterval( dt );






    //  =========  Create a viewer
    //- viewing window
    myViewer = new ExaminerViewer(myWindow);
    myViewer->setSceneGraph(root);
    myViewer->setTitle("My OpenInventor application");
    myViewer->show();
    myViewer->viewAll();
    myViewer->setViewing(false);
    
    //- mouse callbacks 
    eventNode->addEventCallback(SoMouseButtonEvent::getClassTypeId(), myMouseButtonCallback, myViewer);
    eventNode->addEventCallback(SoLocation2Event::getClassTypeId(), myMouseMoveCallback, myViewer);



    //camera
    animation->set_camera(myViewer->getCamera());




    // ====  run the main loop
    cout<<"Type 'i' to show/hide the control window"<<endl;
#ifdef WIN32

    SoWin::show(myWindow);
    SoWin::mainLoop();
#else

    SoQt::show(myWindow);
    SoQt::mainLoop();
#endif


    return EXIT_SUCCESS;
}


